@@ -89,6 +89,7 @@ static void open_filesystem(char *device, int open_flags, blk64_t superblock,
 		current_fs = NULL;
 		return;
 	}
+	current_fs->default_bitmap_type = EXT2FS_BMAP64_RBTREE;
 
 	if (catastrophic)
 		com_err(device, 0, "catastrophic mode - not reading inode or group bitmaps");
@@ -993,29 +994,61 @@ void do_clri(int argc, char *argv[])
 
 void do_freei(int argc, char *argv[])
 {
-	ext2_ino_t inode;
+	unsigned int	len = 1;
+	int		err = 0;
+	ext2_ino_t	inode;
 
-	if (common_inode_args_process(argc, argv, &inode,
-				      CHECK_FS_RW | CHECK_FS_BITMAPS))
+	if (common_args_process(argc, argv, 2, 3, argv[0], "<file> [num]",
+				CHECK_FS_RW | CHECK_FS_BITMAPS))
+		return 1;
+	if (check_fs_read_write(argv[0]))
 		return;
 
-	if (!ext2fs_test_inode_bitmap2(current_fs->inode_map,inode))
+	inode = string_to_inode(argv[1]);
+	if (!inode)
+		return;
+
+	if (argc == 3) {
+		len = parse_ulong(argv[2], argv[0], "length", &err);
+		if (err)
+			return;
+	}
+
+	if (len == 1 &&
+	    !ext2fs_test_inode_bitmap2(current_fs->inode_map,inode))
 		com_err(argv[0], 0, "Warning: inode already clear");
-	ext2fs_unmark_inode_bitmap2(current_fs->inode_map,inode);
+	while (len-- > 0)
+		ext2fs_unmark_inode_bitmap2(current_fs->inode_map, inode);
 	ext2fs_mark_ib_dirty(current_fs);
 }
 
 void do_seti(int argc, char *argv[])
 {
-	ext2_ino_t inode;
+	unsigned int	len = 1;
+	int		err = 0;
+	ext2_ino_t	inode;
 
-	if (common_inode_args_process(argc, argv, &inode,
-				      CHECK_FS_RW | CHECK_FS_BITMAPS))
+	if (common_args_process(argc, argv, 2, 3, argv[0], "<file> [num]",
+				CHECK_FS_RW | CHECK_FS_BITMAPS))
+		return;
+	if (check_fs_read_write(argv[0]))
 		return;
 
-	if (ext2fs_test_inode_bitmap2(current_fs->inode_map,inode))
+	inode = string_to_inode(argv[1]);
+	if (!inode)
+		return;
+
+	if (argc == 3) {
+		len = parse_ulong(argv[2], argv[0], "length", &err);
+		if (err)
+			return;
+	}
+
+	if ((len == 1) &&
+	    ext2fs_test_inode_bitmap2(current_fs->inode_map,inode))
 		com_err(argv[0], 0, "Warning: inode already set");
-	ext2fs_mark_inode_bitmap2(current_fs->inode_map,inode);
+	while (len-- > 0)
+		ext2fs_mark_inode_bitmap2(current_fs->inode_map, inode++);
 	ext2fs_mark_ib_dirty(current_fs);
 }
 #endif /* READ_ONLY */
@@ -1287,35 +1320,6 @@ void do_print_working_directory(int argc, char *argv[])
 }
 
 #ifndef READ_ONLY
-/*
- * Given a mode, return the ext2 file type
- */
-static int ext2_file_type(unsigned int mode)
-{
-	if (LINUX_S_ISREG(mode))
-		return EXT2_FT_REG_FILE;
-
-	if (LINUX_S_ISDIR(mode))
-		return EXT2_FT_DIR;
-
-	if (LINUX_S_ISCHR(mode))
-		return EXT2_FT_CHRDEV;
-
-	if (LINUX_S_ISBLK(mode))
-		return EXT2_FT_BLKDEV;
-
-	if (LINUX_S_ISLNK(mode))
-		return EXT2_FT_SYMLINK;
-
-	if (LINUX_S_ISFIFO(mode))
-		return EXT2_FT_FIFO;
-
-	if (LINUX_S_ISSOCK(mode))
-		return EXT2_FT_SOCK;
-
-	return 0;
-}
-
 static void make_link(char *sourcename, char *destname)
 {
 	ext2_ino_t	ino;
